/*
Copyright (c) 2021 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stddef.h>
#include <string.h>
#include <windows.h>
#include <objbase.h>
#include <mlang.h>
#include "detectcp.h"
#if defined(_MSC_VER)
#  pragma comment(lib, "ole32.lib")
#endif // _MSC_VER

/** \brief Helper implementation which returns the codepage belonging to a Byte Order Mark. Returns 0 if no BOM is found.
 *         Relies on src being at least 4 bytes long. (This has been checked in GetInputCodepage already.)
 * \param src  string of bytes
 * \return Codepage ID belonging to the found Byte Order Mark
 *         0 if no BOM found */
unsigned GuessByBom(const void *src);

/** \brief Helper implementation which informs about a string is likely UTF-32.
 *         Limits the checked number of bytes to 4096 (1024 codepoints) for performance reasons.
 * \param str      string of bytes
 * \param str_len  number of bytes in str, without terminating 0 (if any)
 * \return -1 both ambiguous and double zero bytes found
 *          0 no UTF 32 (odd length, or characters found which are not UTF-32)
 *          1 likely UTF-32 LE
 *          2 likely UTF-32 BE */
int Guess32(const unsigned char *str, size_t str_len);

/** \brief Helper implementation which informs about a string is likely UTF-16.
 *         Uses the IsTextUnicode API function.
 * \param src       string of bytes
 * \param src_size  number of bytes in src, without terminating 0 (if any)
 * \param p_flags   pointer to an int which should be initialized with 0xFFFF, contains the found flags after the function returns
 * \return -1 no UTF-16 but zero bytes found
 *          0 no UTF 16 (odd length, or characters found which are not UTF-16)
 *          1 likely UTF-16 LE
 *          2 likely UTF-16 BE */
int Guess16(const void *src, int src_size, int *p_flags);

/** \brief Helper implementation which informs whether a string is valid ASCII and/or likely UTF-8.
 *         Relies on both str_len being > 3, and str not containing 0 bytes. (These have been checked in GetInputCodepage already.)
 *         Limits the checked number of bytes to 1024 for performance reasons.
 *         May omit up to 3 last bytes from checking to ignore a possibly partial code point at the end of the checked chunk.
 * \param str      string of bytes
 * \param str_len  number of bytes in str, without terminating 0 (if any)
 * \return -1 if suspicious control characters are found (rather binary data)
 *          0 if definitely invalid UTF-8
 *          1 if valid UTF-8, ASCII only
 *          2 if likely UTF-8, containing multibyte sequences that look like UTF-8 */
int Guess8(const unsigned char *str, size_t str_len);

/** \brief Helper implementation which informs whether a string ends with valid ASCII.
 *         Relies on str_len being > 3. (This has been checked in GetInputCodepage already.)
 *         We just check the last few bytes that we didn't evaluate in Guess8 for being ASCII, too.
 *         That's particularly critical for short strings.
 * \param str      string of bytes
 * \param str_len  number of bytes in str, without terminating 0 (if any)
 * \return 1 if the last 4 bytes are all ASCII
 *         0 if one ore more of the last 4 bytes are not ASCII */
int IsTailAscii(const unsigned char *str, size_t str_len);

/** \brief Helper implementation which informs about a string is likely UTF-7.
 *         Relies on a Base64 sequence being finalized with a minus sign (or end of string).
 *         Limits the used number of bytes to 1024 for performance reasons.
 * \param str      string of characters
 * \param str_len  number of characters in str, without terminating 0 (if any)
 * \return 1 if all characters between the first found + and the first following - are valid Base64
 *           or if more than 3 encoded + signs (+- sequences) are found
 *         0 otherwise */
int Guess7(const char *str, size_t str_len);

/** \brief Helper implementation which retrospectively tries to deploy preferred codepages based on the already detected codepage.
 * \param detected_cp   codepage ID reported by DetectInputCodepage
 * \param preferred_cp  codepage preference
 * \return The preferred codepage if it either targets a similar alphabet as the detected codepage, or the
 *          statistical byte distribution is similar even if the charset has completely different semantics.
 *         The previously detected codepage otherwise. */
unsigned AssessUserPreference(unsigned detected_cp, unsigned preferred_cp);

unsigned GetInputCodepage(const void *src, int src_size, unsigned preferred_cp)
{
  if (!src || src_size < 4) // some tests rely on getting at least 4 bytes
    return 0U;

  // try to recognize Byte Order Marks
  const unsigned found = GuessByBom(src);
  if (found)
    return found;

  // try to recognize UTF-32
  const int g32 = Guess32((const unsigned char*)src, (size_t)src_size);
  if (g32 == 1)
    return 12000U; // UTF-32 LE

  if (g32 == 2)
    return 12001U; // UTF-32 BE

  if (g32 == -1)
    return (unsigned)-1; // double zero bytes have been found, however at this point strings should not contain double zero bytes anymore

  // try to recognize UTF-16
  int u16_flags = 0xFFFF;
  const int g16 = Guess16(src, src_size, &u16_flags);
  if (g16 == -1) // likely binary
    return (unsigned)-1;

  if (g16 == 1)
    return 1200U; // UTF-16 LE

  if (g16 == 2)
    return 1201U; // UTF-16 BE

  if (!IsValidCodePage(preferred_cp)) // ensure that we return a valid codepage ID
    preferred_cp = 0U;

  // recognize ASCII and UTF-8
  const int g8 = Guess8((const unsigned char*)src, (size_t)src_size);
  if (g8 == -1) // likely binary
    return (unsigned)-1;

  if (g8 == 2)
    return 65001U; // UTF-8

  if (g8 == 1 && IsTailAscii((const unsigned char*)src, (size_t)src_size)) // characters are ASCII only (however ASCII is a subset of almost every SBCS and DBCS, even Base64 sequences in UTF-7 are ASCII only)
  {
    if (preferred_cp != 0U && preferred_cp != 1200U && preferred_cp != 1201U && preferred_cp != 12000U && preferred_cp != 12001U)
      return preferred_cp; // trust the caller

    // try to recognize UTF-7
    if (Guess7((const char*)src, (size_t)src_size))
      return 65000U; // UTF-7

    return 20127U; // ASCII (falling back to UTF-8 or the default ANSI codepage might be sensible for conversions)
  }

  // use IMultiLanguage2::DetectInputCodepage to get matching ANSI codepages
#if defined(__MINGW32__) // MinGW does not yet implement IID_IMultiLanguage2
#  undef IID_IMultiLanguage2
  const IID IID_IMultiLanguage2 = { 0xDCCFC164U, 0x2B38, 0x11D2, { 0xB7, 0xEC, 0x00, 0xC0, 0x4F, 0x8F, 0x5D, 0x9A } };
#endif
  IMultiLanguage2 *iml2_ptr = NULL;
  const MLDETECTCP hint = (u16_flags & IS_TEXT_UNICODE_DBCS_LEADBYTE) ? MLDETECTCP_DBCS : MLDETECTCP_NONE; // if IsTextUnicode reported found DBCS lead bytes, we'll make use of this hint
  DetectEncodingInfo enc_inf[16] = { 0 };
  int inf_cnt = 16;
  HRESULT yield = E_FAIL;
  if (SUCCEEDED(CoInitialize(NULL)))
  {
    if (SUCCEEDED(CoCreateInstance(&CLSID_CMultiLanguage, NULL, CLSCTX_INPROC_SERVER, &IID_IMultiLanguage2, (void**)&iml2_ptr)))
      yield = iml2_ptr->lpVtbl->DetectInputCodepage(iml2_ptr, hint, (DWORD)preferred_cp, (char*)(intptr_t)src, &src_size, enc_inf, &inf_cnt);

    CoUninitialize();
  }

  if (FAILED(yield) || yield == S_FALSE || inf_cnt < 1)
    return 0U;

  // find the codepage with the best confidence value
  int best_confidence = enc_inf[0].nConfidence;
  unsigned best_cp_match = enc_inf[0].nCodePage;
  for (int i = 1; i < inf_cnt; ++i)
  {
    if (enc_inf[i].nConfidence > best_confidence)
    {
      best_confidence = enc_inf[i].nConfidence;
      best_cp_match = enc_inf[i].nCodePage;
    }
  }

  return best_confidence < 100 // arbitrarily chosen minimum which may need further investigation
    ? 0U // ambiguous
    : AssessUserPreference(best_cp_match, preferred_cp); // trust the caller if reasonable
}

unsigned GuessByBom(const void *src)
{
  if (memcmp(src, "\xFF\xFE\x00\x00", 4U) == 0) // critical order: check UTF-32 LE BOM before checking UTF-16 LE BOM
    return 12000U; // UTF-32 LE BOM found

  if (memcmp(src, "\xFF\xFE", 2U) == 0)
    return 1200U; // UTF-16 LE BOM found

  if (memcmp(src, "\xEF\xBB\xBF", 3U) == 0)
    return 65001U; // UTF-8 BOM found

  if (memcmp(src, "\x84\x31\x95\x33", 4U) == 0)
    return 54936U; // GB-18030 BOM found

  if (memcmp(src, "\xFE\xFF", 2U) == 0)
    return 1201U; // UTF-16 BE BOM found

  if (memcmp(src, "\x00\x00\xFE\xFF", 4U) == 0)
    return 12001U; // UTF-32 BE BOM found

  if (memcmp(src, "+/v", 3U) == 0)
  {
    register const char fourth = *((char*)src + 3);
    if (fourth == '8' || fourth == '9' || fourth == '+' || fourth == '/')
      return 65000U; // UTF-7 BOM found
  }

  return 0U;
}

int Guess32(const unsigned char *str, size_t str_len)
{
  const int is_mltpl4 = !(str_len & 3U); // the length should be a multiple of 4, however we always want to know if quadruple or double zero bytes are found
  str_len = str_len > 4096U ? (size_t)1024U : (str_len >> 2U);

  size_t dblzero_cnt = 0U;
  if (is_mltpl4)
  {
    size_t le_cnt = 0U, be_cnt = 0U;
    for (const unsigned *it = (const unsigned*)str, *const end = it + str_len; it < end; ++it)
    {
      register const unsigned ch = *it;
      if (!ch) // quadruple zero bytes
        return -1;

      dblzero_cnt += !(ch & 0xFFFF0000U) + !(ch & 0x0000FFFFU);
      le_cnt += (ch < 0x0011000U);
      be_cnt += (((ch >> 24U) | ((ch & 0x00FF0000U) >> 8U) | ((ch & 0x0000FF00U) << 8U) | (ch << 24U)) < 0x0011000U);
    }

    if (le_cnt == str_len)
      return 1; // UTF-32 LE

    if (be_cnt == str_len)
      return 2; // UTF-32 BE
  }
  else
  {
    for (const unsigned *it = (const unsigned*)str, *const end = it + str_len; it < end; ++it)
    {
      register const unsigned ch = *it;
      if (!ch) // quadruple zero bytes
        return -1;

      dblzero_cnt += !(ch & 0xFFFF0000U) + !(ch & 0x0000FFFFU);
    }
  }

  return dblzero_cnt ? -1 : 0; // consider double zero bytes as break condition and refrain from further checks in GetInputCodepage
}

int Guess16(const void *src, int src_size, int *p_flags)
{
  IsTextUnicode(src, src_size, p_flags);
  const int le_flags = *p_flags & IS_TEXT_UNICODE_UNICODE_MASK, be_flags = *p_flags & IS_TEXT_UNICODE_REVERSE_MASK;
  if ((le_flags || be_flags) && !(*p_flags & IS_TEXT_UNICODE_ODD_LENGTH) && (*p_flags & IS_TEXT_UNICODE_NULL_BYTES)) // the latter prevents the "Bush hid the facts" bug (we always expect to see 0 bytes, at least for spaces between words)
  {
    if ((le_flags && be_flags) || (*p_flags & IS_TEXT_UNICODE_NOT_UNICODE_MASK)) // ambiguous
      return -1;

    if (le_flags)
      return 1; // UTF-16 LE

    if (be_flags)
      return 2; // UTF-16 BE
  }

  if (*p_flags & IS_TEXT_UNICODE_NULL_BYTES) // at this point strings should not contain any 0 bytes
    return -1;

  return 0;
}

int Guess8(const unsigned char *str, size_t str_len)
{
  static const char lengths[] = { 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 2, 2, 3, 3, 4, 8 }; // based on Christopher Wellons' lengths array, Public Domain, https://github.com/skeeto/branchless-utf8
  static const char shift_r[] = { 0, 9, 6, 3, 0 };
  if (str_len > 1024U)
    str_len = 1024U; // upper limit of checked code units

  int seq = 0; // 1 if multibyte sequences are found
  for (const unsigned char *it = str, *const end = str + str_len - 3; it < end;)
  {
    register const unsigned char byte_val = *it;
    register const int char_len1 = lengths[byte_val >> 3];
    const int char_len2 = lengths[it[1] >> 3], char_len3 = lengths[it[2] >> 3], char_len4 = lengths[it[3] >> 3];
    // check if we got the right number of continuation bytes
    if (!char_len1 || ((char_len2 << 6) | (char_len3 << 3) | char_len4) >> shift_r[char_len1] || (char_len1 | char_len2 | char_len3 | char_len4) >> 3)
      return 0;

    it += char_len1;
    if (char_len1 == 1) // 01..7F is ASCII
    {
      if (byte_val < 0x20U && (byte_val < 0x08U || byte_val > 0x0DU)) // a lot of the C0 control characters are rather suspicious in text files
        return -1;

      continue;
    }

    seq = 1;
  }

  return seq + 1;
}

int IsTailAscii(const unsigned char *str, size_t str_len)
{
  for (const unsigned char *const end = str + str_len, *it = end - 4; it < end; ++it)
    if (*it > 0x7FU)
      return 0;

  return 1;
}

int Guess7(const char *str, size_t str_len)
{
  int truncated = 0;
  if (str_len > 1024U)
  {
    truncated = 1;
    str_len = 1024U; // upper limit of used code units
  }

  int confidence = 0;
  for (const char *it = str, *const end = str + str_len; it < end;)
  {
    const char *const plus_ptr = memchr(it, '+', (size_t)(end - it));
    if (!plus_ptr)
      break;

    int ignore_mltpl3 = truncated;
    const char *minus_ptr = memchr(plus_ptr, '-', (size_t)(end - plus_ptr));
    if (minus_ptr)
      ignore_mltpl3 = 0;
    else
      minus_ptr = end;

    it = plus_ptr + 1;
    if (it == minus_ptr) // + is encoded +-
    {
      ++confidence;
      continue;
    }

    // the number of bytes in the Base64 sequence must be a multiple of 3
    if (!ignore_mltpl3 && (minus_ptr - it) % 3)
      return (confidence > 3);

    // the sequence must consist of valid Base64 characters
    for (; it < minus_ptr; ++it)
    {
      register const char ch = *it;
      if (!(ch == '+' || (ch >= '/' && ch <= '9') || (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z')))
        return (confidence > 3);
    }

    return 1;
  }

  return (confidence > 3);
}

unsigned AssessUserPreference(unsigned detected_cp, unsigned preferred_cp)
{
  if (preferred_cp && (
      (detected_cp == 932U && preferred_cp == 874U) ||
      (detected_cp == 936U && (preferred_cp == 20936U || preferred_cp == 54936U)) ||
      (detected_cp == 1250U && preferred_cp == 28592U) ||
      (detected_cp == 1252U && (preferred_cp == 437U || preferred_cp == 850U || preferred_cp == 860U || preferred_cp == 861U || preferred_cp == 863U || preferred_cp == 865U || preferred_cp == 28591U)) ||
      (detected_cp == 1253U && preferred_cp == 28597U) ||
      (detected_cp == 1254U && (preferred_cp == 857U || preferred_cp == 28599U)) ||
      (detected_cp == 1255U && preferred_cp == 28598U) ||
      (detected_cp == 1257U && (preferred_cp == 775U || preferred_cp == 28594U)) ||
      (detected_cp == 51932U && preferred_cp == 20932U)))
    return preferred_cp;

  return detected_cp;
}
