/*
Copyright (c) 2021 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#define _UCRT
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <share.h>
#include <windows.h>
#include <shellapi.h>
#include "detectcp.h"
#if defined(_MSC_VER)
#  pragma comment(lib, "shell32.lib")
#endif // _MSC_VER

// Defining this macro will only affect the GCC compiler and can reduce the size of the utility. As soon as you define it you have to add these linker options:
//   -Wl,-e,__start -nostartfiles
#define USE_START

#if defined(__GNUC__) && defined(USE_START)
# define QUIT(_status) exit(_status)
# ifdef __i386
void _start(void);
void _start(void)
# else
void __start(void);
void __start(void)
# endif
#else
# define QUIT(_status) return (_status)
int main(void)
#endif
{
  int argc;
  wchar_t **argv = CommandLineToArgvW(GetCommandLineW(), &argc);
  if (argc < 2 || wcscmp(argv[1], L"/?") == 0)
  {
    LocalFree(argv);
    puts("Attempts to determine the encoding of a text file.\n"
         " NOTE: Do not rely on the reported codepage ID being correct.\n\n"
         "Syntax:\n"
         "DetectCodepage file [preference]\n"
         "  file        Name of the text file to be examined.\n"
         "  preference  (optional) Codepage ID to be preferred.\n\n"
         "If the encoding is successfully estimated, the related codepage\n"
         " ID will be printed.\n"
         "If the encoding is indefinite, 0 will be printed.\n"
         "If the data is rather binary, -1 will be printed.\n\n");
    QUIT(0);
  }

  FILE *stream = _wfsopen(argv[1], L"rb", _SH_DENYWR);
  if (!stream)
  {
    LocalFree(argv);
    QUIT(1);
  }

  unsigned char str[4096];
  const int str_cnt = (int)fread(str, 1U, 4096U, stream);
  fclose(stream);

  unsigned preferred_cp = 0U;
  if (argc > 2)
  {
    wchar_t *endptr = NULL;
    errno = 0;
    const long cp = wcstol(argv[2], &endptr, 10);
    if (!errno && !*endptr && cp > -1L)
    {
      if (cp == 0L)
        preferred_cp = GetACP();
      else if (cp == 1L)
        preferred_cp = GetOEMCP();
      else
        preferred_cp = (const unsigned)cp;
    }
  }

  LocalFree(argv);
  printf("%d\n", (int)GetInputCodepage(str, str_cnt, preferred_cp));
  QUIT(0);
}
